/*
  File name: CSDLClass.cpp
  Copyright: (C) 2005 - 2007 Tomasz Kazmierczak

  Creation date: 01.07.2005 12:51
  Last modification date: 24.04.2008

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

  SDL interface
*/

#include <memory.h>
#include "CSDLClass.h"
#include "LoopCtrl.h"

CSDLClass *CSDLClass::instance()
{
	static CSDLClass inst;
	return &inst;
}

//init/free functions
CSDLClass::CSDLClass() : mScreen(0), mJoystick(0), mMusic(0), mAudio(false),
	mLoadedSoundsCount(0), mMenuMouseXPos(0), mMenuMouseYPos(0),
	mLastMouseInputTime(0), mLastJoyInputTime(0), mScreenSizesCount(0),
	mScreenSizes(0)
{
	memset(mSounds, 0, sizeof(mSounds));
}

bool CSDLClass::init()
{
	if(-1 == SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK | SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_NOPARACHUTE)) {
		fprintf(stderr, "!SDL_Init(): %s\n", SDL_GetError());
		return false;
	}

	//check for needed screen sizes
	//size for 2D video mode
	uint32_t uiFlags = SDL_HWSURFACE | SDL_FULLSCREEN | SDL_DOUBLEBUF;
	if(not SDL_VideoModeOK(640, 480, 24, uiFlags)) {
		fprintf(stderr, "Hardware surface in 640x480 mode not available!\n");
		SDL_Quit();
		return false;
	}
	//sizes for OpenGL (we need at least one of them)
	uiFlags &= ~SDL_DOUBLEBUF;
	uiFlags |= SDL_OPENGL;
	static ScreenSize sizes[] = {{{CSDL_DEFAULT_SCREEN_W_TXT, CSDL_DEFAULT_SCREEN_H_TXT},
								   {CSDL_DEFAULT_SCREEN_W, CSDL_DEFAULT_SCREEN_H}, true},
								  {{"800", "600"}, {800, 600}, true},
								  {{"1024", "768"}, {1024, 768}, true},
								  {{"1152", "864"}, {1152, 864}, true},
								  {{"1280", "800"}, {1280, 800}, true},
								  {{"1280", "960"}, {1280, 960}, true},
								  {{"1280", "1024"}, {1280, 1024}, true},
								  {{"1366", "768"}, {1366, 768}, true},
								  {{"1600", "1024"}, {1600, 1024}, true},
								  {{"1600", "1200"}, {1600, 1200}, true},
								  {{"1920", "1200"}, {1920, 1200}, true},
								  {{"2048", "1536"}, {2048, 1536}, true}};
	bool anySizeAvailable = false;
	//how many sizes the ScreenSizes array consists of?
	mScreenSizesCount = sizeof(sizes) / sizeof(ScreenSize);
	for(int32_t i = 0; i < mScreenSizesCount; i++) {
		if(SDL_VideoModeOK(sizes[i].resolution[0], sizes[i].resolution[1], 24, uiFlags) < 16)
			sizes[i].available = false;
		else anySizeAvailable = true;
	}
	if(not anySizeAvailable) {
		fprintf(stderr, "No required screen resolutions in OpenGL mode available!\n"
						"At least one of the following screen resolutions in OpenGL\n"
						"mode is needed to run the game (in at least 16 bit color\n"
						"mode):\n");
		for(int32_t i = 0; i < mScreenSizesCount; i++)
			fprintf(stderr, "%dx%d ", sizes[i].resolution[0], sizes[i].resolution[1]);
		fprintf(stderr, "\n");
		SDL_Quit();
		return false;
	}
	mScreenSizes = sizes;

#if not defined(DEBUG)
 	SDL_ShowCursor(SDL_DISABLE);
 	SDL_WM_GrabInput(SDL_GRAB_ON);
#endif

	//if there is a joystick then we enable it's events and open it
	mJoystick = 0;
	if(SDL_NumJoysticks())
	{
		SDL_JoystickEventState(SDL_ENABLE);
		mJoystick = SDL_JoystickOpen(0);
	}

	//audio initialization
	if(0 == Mix_OpenAudio(44100, AUDIO_S16, 2, 4096))
		mAudio = true;
	else
		fprintf(stderr, "!Mix_OpenAudio(44100, AUDIO_S16, 2, 4096): %s\n", Mix_GetError());

	mMenuMouseXPos = 0;
	mMenuMouseYPos = 0;
	return true;
}

bool CSDLClass::switchVideoMode(int32_t iWidth, int32_t iHeight, VideoMode eMode)
{
#if defined(DEBUG)
	uint32_t dwFlags = SDL_HWSURFACE;
#else
	uint32_t dwFlags = SDL_HWSURFACE | SDL_FULLSCREEN;
#endif

#if defined(__WIN32__)
	uint32_t uiBits;
#endif
	switch(eMode)
	{
	case MODE_2D:
		{
			dwFlags |= SDL_DOUBLEBUF;
#if defined(__WIN32__)
			uiBits = 16;
#endif
		}
		break;
	case MODE_3D:
		{
			//attribs for OGL
			SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
			SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
			SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 1);
			SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE, 8);
			SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE, 8);
			SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE, 8);
			dwFlags |= SDL_OPENGL;
#if defined(__WIN32__)
			uiBits = 24;
#endif
		}
		break;
	}

#if defined(__WIN32__)
	mScreen = SDL_SetVideoMode(iWidth, iHeight, uiBits, dwFlags);
#else
	mScreen = SDL_SetVideoMode(iWidth, iHeight, 24, dwFlags);
#endif

	if(not mScreen)
	{
		fprintf(stderr, "!SDL_SetVideoMode(%d, %d, %s): %s\n",
						iWidth, iHeight, (dwFlags & SDL_OPENGL) ? "3D" : "2D",
						SDL_GetError());
		return false;
	}
	return true;
}

void CSDLClass::free()
{
	freeSounds();
	mImgArray.freeIA();
	Mix_CloseAudio();
	if(mJoystick)
		SDL_JoystickClose(mJoystick);
	SDL_WM_GrabInput(SDL_GRAB_OFF);
	SDL_ShowCursor(SDL_ENABLE);
	SDL_Quit();
}

//input functions
void CSDLClass::delayInput()
{
	uint32_t dwTime = SDL_GetTicks();
	mLastMouseInputTime = dwTime;
	mLastJoyInputTime = dwTime;
}

void CSDLClass::clearEventQueue()
{
	SDL_Event sEvent;
	while(SDL_PollEvent(&sEvent));
}

SDLKey CSDLClass::readInputMenu()
{
	uint32_t dwTime = SDL_GetTicks();
	SDL_Event sEvent;
	//check if there are any events
	while(SDL_PollEvent(&sEvent))
	{
		switch(sEvent.type)
		{
		case SDL_KEYDOWN:
			{
				switch(sEvent.key.keysym.sym)
				{
				case SDLK_KP_ENTER:
					return SDLK_RETURN;
				case SDLK_ESCAPE:
				case SDLK_RETURN:
				case SDLK_SPACE:
				case SDLK_UP:
				case SDLK_DOWN:
				case SDLK_LEFT:
				case SDLK_RIGHT:
					return sEvent.key.keysym.sym;
				default:
					break;
				}
			}
			break;

		case SDL_MOUSEMOTION:
			{
				mMenuMouseYPos += sEvent.motion.yrel;
				if(mMenuMouseYPos > 30)
				{
					mMenuMouseYPos = 0;
					return SDLK_DOWN;
				}
				if(mMenuMouseYPos < -30)
				{
					mMenuMouseYPos = 0;
					return SDLK_UP;
				}
				mMenuMouseXPos += sEvent.motion.xrel;
				if(mMenuMouseXPos > 30)
				{
					mMenuMouseXPos = 0;
					return SDLK_RIGHT;
				}
				if(mMenuMouseXPos < -30)
				{
					mMenuMouseXPos = 0;
					return SDLK_LEFT;
				}
			}
			break;

		default:
			break;
		}
	}
	//get mouse input
	if(SDL_GetMouseState(0, 0) & SDL_BUTTON(SDL_BUTTON_LEFT) or
		SDL_GetMouseState(0, 0) & SDL_BUTTON(SDL_BUTTON_RIGHT))
	{
		if(dwTime - mLastMouseInputTime < 250)
			return SDLK_UNKNOWN;
		mLastMouseInputTime = dwTime;
		return SDLK_RETURN;
	}
	//if mouse buttons are released we don't delay their next input
	else
	    mLastMouseInputTime = dwTime - 250;
	//get joystick input
	if(mJoystick)
	{
		int16_t siY = SDL_JoystickGetAxis(mJoystick, 1);
		bool bJoyBtn = (bool)SDL_JoystickGetButton(mJoystick, 0);
		//if there is no input from joy we don't delay its next input
		if(siY < 3200 and siY > -3200 and not bJoyBtn)
		{
			mLastJoyInputTime = dwTime - 250;
			return SDLK_UNKNOWN;
		}

		//delay of movement and 'fire' repetition
		if(dwTime - mLastJoyInputTime < 250)
			return SDLK_UNKNOWN;
		mLastJoyInputTime = dwTime;
		//button
		if(bJoyBtn)
			return SDLK_RETURN;
		//axes
		if(siY > 3200)
			return SDLK_DOWN;
		if(siY < -3200)
			return SDLK_UP;
	}
	return SDLK_UNKNOWN;
}

void CSDLClass::readInputGame(InputArray *pInputArray)
{
	SDL_Event sEvent;
	//check if there are any events
	uint32_t dwReadCounter = 0;
	while(SDL_PollEvent(&sEvent))
	{
		switch(sEvent.type)
		{
			case SDL_KEYDOWN:
			{
				if(dwReadCounter == 5)
					break;
				pInputArray->eKey[dwReadCounter] = sEvent.key.keysym.sym;
				pInputArray->bStatus[dwReadCounter] = true;
				++dwReadCounter;
			}
			break;
			case SDL_KEYUP:
			{
				if(dwReadCounter == 5)
					break;
				pInputArray->eKey[dwReadCounter] = sEvent.key.keysym.sym;
				pInputArray->bStatus[dwReadCounter] = false;
				++dwReadCounter;
			}
			break;
			case SDL_MOUSEMOTION:
			{
				pInputArray->siMouseX = sEvent.motion.xrel;
				pInputArray->siMouseY = sEvent.motion.yrel;
			}
			break;
			default:
				break;
		}
	}
	pInputArray->pKeyMap = SDL_GetKeyState(0);
	//if left mouse button is pressed
	if(SDL_GetMouseState(0, 0) & SDL_BUTTON(SDL_BUTTON_LEFT))
		pInputArray->bLMB = true;
	else
		pInputArray->bLMB = false;
	if(SDL_GetMouseState(0, 0) & SDL_BUTTON(SDL_BUTTON_RIGHT))
		pInputArray->bRMB = true;
	else
		pInputArray->bRMB = false;
	//get joystick input
	if(mJoystick)
	{
		//axes
		int16_t siX = SDL_JoystickGetAxis(mJoystick, 0);
		int16_t siY = SDL_JoystickGetAxis(mJoystick, 1);
		if(siY < -3200)
			pInputArray->bJoyUp = true;
		else
			pInputArray->bJoyUp = false;
		if(siY > 3200)
			pInputArray->bJoyDown = true;
		else
			pInputArray->bJoyDown = false;
		if(siX > 3200)
			pInputArray->bJoyRight = true;
		else
			pInputArray->bJoyRight = false;
		if(siX < -3200)
			pInputArray->bJoyLeft = true;
		else
			pInputArray->bJoyLeft = false;
		//button
		pInputArray->bJBtn = (bool)SDL_JoystickGetButton(mJoystick, 0);
	}
}

//display functions
void CSDLClass::displayImage(uint32_t dwImgNum, SDL_Rect *pSrcRect, int16_t iX, int16_t iY)
{
	if(dwImgNum >= mImgArray.mImagesCount)
		return;
	SDL_Rect sDestRect = {iX, iY, mImgArray.mImageDims[dwImgNum].dwWidth,
							mImgArray.mImageDims[dwImgNum].dwHeight};
	SDL_BlitSurface(mImgArray.mImageSurface[dwImgNum], pSrcRect, mScreen, &sDestRect);
}

void CSDLClass::displayText(uint32_t dwImgNum, const char *szText, int16_t iX, int16_t iY)
{
	if(dwImgNum >= mImgArray.mImagesCount)
		return;
	uint16_t wWidth = mImgArray.mImageDims[dwImgNum].dwWidth / 16;
	uint16_t wHeight = mImgArray.mImageDims[dwImgNum].dwHeight / 16;
	SDL_Rect sSrcRect = {0, 0, wWidth, wHeight};
	SDL_Rect sDestRect = {iX, iY, wWidth, wHeight};
	uint32_t dwTextLen = strlen(szText);
	int32_t iNewLinesCount = 0;
	for(uint32_t i = 0; i < dwTextLen; ++i)
	{
		uint8_t c = szText[i];
		//if new line
		if(c == '\n')
		{
			sDestRect.x = iX;
			iNewLinesCount++;
			continue;
		}
		//find an apropirate character in the font image
		uint16_t wXOffs = c % 16;
		uint16_t wYOffs = c / 16;
		sSrcRect.x = wXOffs * wWidth;
		sSrcRect.y = wYOffs * wHeight;
		//we must modify the sDestRect.y every time before blitting because
		//SDL_BlitSurface() saves the final blit rectangle (after clipping)
		//in sDestRect, so if the Y coord was, for example, negative (text was
		//scrolled above the upper boundary of the screen) it has been changed
		//to 0 and the next drawn characters (in current line and in next lines
		//also) would not have been scrolled up
		//if we wanted to scroll the text horizontally we would do the same
		//with the X coord, of course
		sDestRect.y = iY + iNewLinesCount * (wHeight * 3 / 2);
		//blit the character image to the screen
		SDL_BlitSurface(mImgArray.mImageSurface[dwImgNum], &sSrcRect, mScreen, &sDestRect);
		sDestRect.x += wWidth;
	}
}

void CSDLClass::blackScreen()
{
	SDL_FillRect(mScreen, 0, 0);
}

int32_t CSDLClass::findFirstAvailableScreenSize(int32_t start)
{
	//NOTE: if the CSDLClass was initialized in spite of
	//a failure to find at least one valid screen mode
	//this function will return iStart and will _not_
	//report the lack of screen modes
	int32_t i = start + 1;
	for(; i < mScreenSizesCount; ++i) {
		if(mScreenSizes[i].available)
			break;
	}
	if(i < mScreenSizesCount)
		return i;
	for(i = 0; i < start; ++i) {
		if(mScreenSizes[i].available)
			break;
	}
	return i;
}

//audio functions
bool CSDLClass::loadMusic(const char *szFileName)
{
	if(not mAudio)
		return true;

	if(mMusic)
		Mix_FreeMusic(mMusic);

	mMusic = Mix_LoadMUS(szFileName);
	if(not mMusic)
	{
		fprintf(stderr, "!Mix_LoadMUS(\"%s\"): %s\n", szFileName, Mix_GetError());
		return false;
	}
	return true;
}

void CSDLClass::freeMusic()
{
	if(not mAudio)
		return;
	Mix_FreeMusic(mMusic);
	mMusic = 0;
}

void CSDLClass::playMusic(int32_t iLoops)
{
	if(mAudio and -1 == Mix_PlayMusic(mMusic, iLoops))
		fprintf(stderr, "!Mix_PlayMusic(): %s\n", Mix_GetError());
}

void CSDLClass::stopMusic()
{
	Mix_HaltMusic();
}

int32_t CSDLClass::loadSound(const char *szFileName)
{
	if(not mAudio)
		return CSDL_AUDIO_NOT_INITIALIZED;

	if(mLoadedSoundsCount == MAX_SOUNDS_COUNT) {
		fprintf(stderr, "Can't load more sounds than %d!\n", MAX_SOUNDS_COUNT);
		return CSDL_LOAD_SOUND_FAILED;
	}
	mSounds[mLoadedSoundsCount] = Mix_LoadWAV(szFileName);
	if(not mSounds[mLoadedSoundsCount]) {
		fprintf(stderr, "!Mix_LoadWAV(\"%s\"): %s\n", szFileName, Mix_GetError());
		return CSDL_LOAD_SOUND_FAILED;
	}
	mLoadedSoundsCount++;
	return (mLoadedSoundsCount-1);
}

void CSDLClass::freeSounds()
{
	//stop playback on all channels
	Mix_HaltChannel(-1);
	//free sounds
	for(int32_t i = 0; i < mLoadedSoundsCount; ++i) {
		Mix_FreeChunk(mSounds[i]);
		mSounds[i] = 0;
	}
	mLoadedSoundsCount = 0;
}

void CSDLClass::playSoundFadeIn(int32_t sound, int32_t channel, int32_t loops, int32_t fadeInTime)
{
	if(sound < 0)
		return;
	Mix_Volume(sound, 128);
	if(mAudio and -1 == Mix_FadeInChannel(channel, mSounds[sound], loops, fadeInTime))
		fprintf(stderr, "!Mix_FadeInChannel(): %s\n", Mix_GetError());
}

void CSDLClass::playSound(int32_t sound, int32_t channel, int32_t volume, int32_t loops)
{
	if(sound < 0)
		return;
	Mix_Volume(channel, volume);
	if(mAudio and -1 == Mix_PlayChannel(channel, mSounds[sound], loops))
		fprintf(stderr, "!Mix_PlayChannel(): %s\n", Mix_GetError());
}

void CSDLClass::setSoundVolume(int32_t channel, int32_t volume)
{
	if(not mAudio)
		return;
	Mix_Volume(channel, volume);
}

void CSDLClass::stopSound(int32_t channel, int32_t iFadeOutTime)
{
	Mix_FadeOutChannel(channel, iFadeOutTime);
}

void CImagesArray::freeIA()
{
	if(mImageDims)
	{
		delete[] mImageDims;
		mImageDims = 0;
	}

	if(mImageSurface)
	{
		for(uint32_t i = 0; i < mImagesCount; i++)
		{
			if(mImageSurface[i])
				SDL_FreeSurface(mImageSurface[i]);
		}
		delete[] mImageSurface;
		mImageSurface = 0;
	}
	mImagesCount = 0;
}

bool CImagesArray::loadImg(const char* szFarFileName, const char *szImgFileName,  uint32_t *pImgNum)
{
	//create new arrays
	SDL_Surface **pTmpSurf = new SDL_Surface*[mImagesCount + 1];
	if(pTmpSurf == 0) {
		fprintf(stderr, "CImagesArray::loadImg(%s, %s): !pTmpSurf\n", szFarFileName, szImgFileName);
		return false;
	}
	ImgDims *pTmpDims = new ImgDims[mImagesCount + 1];
	if(pTmpDims == 0) {
		delete[] pTmpSurf;
		fprintf(stderr, "CImagesArray::loadImg(%s, %s): !pTmpDims\n", szFarFileName, szImgFileName);
		return false;
	}
	//copy elements (pointers to SDL_Surfaces and image dimms structres)
	//from old arrays to new ones
	for(uint32_t i = 0; i < mImagesCount; i++) {
		pTmpSurf[i] = mImageSurface[i];
		pTmpDims[i] = mImageDims[i];
	}

	//read image data
	uint8_t *pData = farReadImgData(szFarFileName, szImgFileName, pTmpDims[mImagesCount]);
	if(pData == 0) {
		delete[] pTmpDims;
		delete[] pTmpSurf;
		fprintf(stderr,"CImagesArray::loadImg(): !farReadImgData(%s, %s)\n", szFarFileName, szImgFileName);
		return false;
	}
	uint32_t w = pTmpDims[mImagesCount].dwWidth;
	uint32_t h = pTmpDims[mImagesCount].dwHeight;
	uint32_t bpp = pTmpDims[mImagesCount].dwBPP;
	uint32_t size = pTmpDims[mImagesCount].dwSize;

	uint32_t dwAlphaMask = 0;
	if(bpp == 32)
		dwAlphaMask = 0xFF000000;
	//create the surface (in video memory - SDL_HWSURFACE)
	pTmpSurf[mImagesCount] = SDL_CreateRGBSurface(SDL_HWSURFACE, w, h, bpp,
									0x00FF0000, 0x0000FF00, 0x000000FF, dwAlphaMask);
	if(not pTmpSurf[mImagesCount]) {
		delete[] pTmpDims;
		delete[] pTmpSurf;
		delete[] pData;
		fprintf(stderr,"CImagesArray::loadImg(%s, %s): !pTmpSurf[mImagesCount]\n",
						szFarFileName, szImgFileName);
		return false;
	}
	//copy pixels
	uint32_t linewidth = w * (bpp / 8);  //bytes per line
	//swap vertically while copying
	uint8_t *dest = (uint8_t *)(pTmpSurf[mImagesCount]->pixels);
	for(uint32_t i = 0; i < h; ++i)
		memcpy(&dest[i * linewidth], &pData[size - ((i + 1) * linewidth)], linewidth);
	//we don't need the data any more
	delete[] pData;

	//delete old arrays and copy pointers to new arrays
	delete[] mImageDims;
	delete[] mImageSurface;
	mImageDims = pTmpDims;
	mImageSurface = pTmpSurf;
	*pImgNum = mImagesCount;
	mImagesCount++;
	return true;
}

const ImgDims *CImagesArray::getImageDimensions(uint32_t uiImageNum)
{
	if(uiImageNum < mImagesCount)
		return &mImageDims[uiImageNum];
	return 0;
}
